<?php

namespace InSegment\ApiCore\Traits;

use Illuminate\Support\Str;
use Illuminate\Database\Eloquent\Model;
use InSegment\ApiCore\Services\AttributeCache;

/**
 * To be used with \Illuminate\Database\Eloquent\Model
 */
trait CachesAttributes
{
    private $attributeCacheKey = '';
    
    private function getCachedAttributeCharactestics(string $key)
    {
        if (($charx = AttributeCache::try(static::class, $this->attributeCacheKey, $key)) === null) {
            $charx =
                ($this->hasCast($key) ? AttributeCache::CACHE_HAS_CAST : 0) |
                (method_exists($this, $key) ? AttributeCache::CACHE_HAS_METHOD : 0) |
                (method_exists(Model::class, $key) ? AttributeCache::CACHE_BASE_HAS_METHOD : 0) |
                ($this->hasGetMutator($key) ? AttributeCache::CACHE_HAS_GET_MUTATOR : 0) |
                ($this->hasSetMutator($key) ? AttributeCache::CACHE_HAS_SET_MUTATOR : 0) |
                ($this->isDateAttribute($key) ? AttributeCache::CACHE_IS_DATE_ATTRIBUTE : 0) |
                ($this->isJsonCastable($key) ? AttributeCache::CACHE_IS_JSON_CASTABLE : 0) |
                (Str::contains($key, '->') ? AttributeCache::CACHE_FILL_JSON : 0) |
                (in_array($key, $this->getDates()) ? AttributeCache::CACHE_IN_DATES : 0);
                    
            AttributeCache::cache(static::class, $this->attributeCacheKey, $key, $charx);
        }
        
        return $charx;
    }
    
    /**
     * Get an attribute from the model.
     *
     * @param  string  $key
     * @return mixed
     */
    public function getAttribute($key)
    {
        if (! $key) {
            return;
        }
        
        $charx = $this->getCachedAttributeCharactestics($key);

        // If the attribute exists in the attribute array or has a "get" mutator we will
        // get the attribute's value. Otherwise, we will proceed as if the developers
        // are asking for a relationship's value. This covers both types of values.
        if (array_key_exists($key, $this->attributes) || $charx & AttributeCache::CACHE_HAS_GET_MUTATOR) {
            return $this->getAttributeValue($key);
        }

        // Here we will determine if the model base class itself contains this given key
        // since we do not want to treat any of those methods are relationships since
        // they are all intended as helper methods and none of these are relations.
        if ($charx & AttributeCache::CACHE_BASE_HAS_METHOD) {
            return;
        }

        return $this->getRelationValue($key);
    }

    /**
     * Get a plain attribute (not a relationship).
     *
     * @param  string  $key
     * @return mixed
     */
    public function getAttributeValue($key)
    {
        $charx = $this->getCachedAttributeCharactestics($key);
        $value = $this->getAttributeFromArray($key);

        // If the attribute has a get mutator, we will call that then return what
        // it returns as the value, which is useful for transforming values on
        // retrieval from the model to a form that is more useful for usage.
        if ($charx & AttributeCache::CACHE_HAS_GET_MUTATOR) {
            return $this->mutateAttribute($key, $value);
        }

        // If the attribute exists within the cast array, we will convert it to
        // an appropriate native PHP type dependant upon the associated value
        // given with the key in the pair. Dayle made this comment line up.
        if ($charx & AttributeCache::CACHE_HAS_CAST) {
            return $this->castAttribute($key, $value);
        }

        // If the attribute is listed as a date, we will convert it to a DateTime
        // instance on retrieval, which makes it quite convenient to work with
        // date fields without having to create a mutator for each property.
        if ($charx & AttributeCache::CACHE_IN_DATES && $value !== null) {
            return $this->asDateTime($value);
        }

        return $value;
    }

    /**
     * Get a relationship.
     *
     * @param  string  $key
     * @return mixed
     */
    public function getRelationValue($key)
    {
        $charx = $this->getCachedAttributeCharactestics($key);
        
        // If the key already exists in the relationships array, it just means the
        // relationship has already been loaded, so we'll just return it out of
        // here because there is no need to query within the relations twice.
        if ($this->relationLoaded($key)) {
            return $this->relations[$key];
        }

        // If the "attribute" exists as a method on the model, we will just assume
        // it is a relationship and will load and return results from the query
        // and hydrate the relationship's value on the "relationships" array.
        if ($charx & AttributeCache::CACHE_HAS_METHOD) {
            return $this->getRelationshipFromMethod($key);
        }
    }

    /**
     * Set a given attribute on the model.
     *
     * @param  string  $key
     * @param  mixed  $value
     * @return $this
     */
    public function setAttribute($key, $value)
    {
        $charx = $this->getCachedAttributeCharactestics($key);
        
        // First we will check for the presence of a mutator for the set operation
        // which simply lets the developers tweak the attribute as it is set on
        // the model, such as "json_encoding" an listing of data for storage.
        if ($charx & AttributeCache::CACHE_HAS_SET_MUTATOR) {
            return $this->{'set'.Str::studly($key).'Attribute'}($value);
        }

        // If an attribute is listed as a "date", we'll convert it from a DateTime
        // instance into a form proper for storage on the database tables using
        // the connection grammar's date format. We will auto set the values.
        elseif ($charx & AttributeCache::CACHE_IS_DATE_ATTRIBUTE && $value) {
            $value = $this->fromDateTime($value);
        }

        if ($charx & AttributeCache::CACHE_IS_JSON_CASTABLE && $value !== null) {
            $value = $this->castAttributeAsJson($key, $value);
        }

        // If this attribute contains a JSON ->, we'll set the proper value in the
        // attribute's underlying array. This takes care of properly nesting an
        // attribute in the array's value in the case of deeply nested items.
        if ($charx & AttributeCache::CACHE_FILL_JSON) {
            return $this->fillJsonAttribute($key, $value);
        }

        $this->attributes[$key] = $value;

        return $this;
    }
    
    /**
     * Set the value of the "created at" attribute.
     *
     * @param  mixed  $value
     * @return $this
     */
    public function setCreatedAt($value)
    {
        $this->setAttribute(static::CREATED_AT, $value);

        return $this;
    }

    /**
     * Set the value of the "updated at" attribute.
     *
     * @param  mixed  $value
     * @return $this
     */
    public function setUpdatedAt($value)
    {
        $this->setAttribute(static::UPDATED_AT, $value);

        return $this;
    }

    /**
     * Get the format for database stored dates.
     *
     * @return string
     */
    protected function getDateFormat()
    {
        if (!$this->dateFormat) {
            $connectionName = $this->getConnectionName();
            if (($format = AttributeCache::tryDateFormat($connectionName)) === null) {
                $format = $this->getConnection()->getQueryGrammar()->getDateFormat();
                AttributeCache::cacheDateFormat($format, $connectionName);
            }
            
            return $format;
        }
        
        return $this->dateFormat;
    }

}
